#include <cassert>
#include <string>
#include <ctime>
#include <cmath>
#include <iomanip>
#include <limits>
#include "EMD_header.h"

using namespace std;



#define WARP_SIZE 32


//CUDA kernels
__global__ void mUpdate(float* d_Phi, float* d_m, float* d_m_temp, int n, float dx, float mu);
__global__ void rhoStep(float* d_Phi, float* d_t_rho0, float* d_t_rho0_temp, float* d_t_rho1, float* d_t_rho1_temp, int nu, int n);
__global__ void PhiUpdate(float* d_Phi, float *d_m_temp, float* d_t_rho0_temp, float* d_t_rho1_temp, int n, float dx, float tau);
__global__ void cuda_parallel_minmax(float *rho0, int num_elements, float *max, float *min);
__global__ void cuda_parallel_min(float *rho0, int num_elements, float *min);
__global__ void cuda_parallel_max(float *rho0, int num_elements, float *max);
__global__ void parallel_sum(float *rho, float* t_rho, int num_elements, float theta_mid, float *sum);
__global__ void rho_commit(float *rho, float* t_rho, float* t_rho_temp, int num_elements, float theta_mid);


void err_chk(cudaError err) {
  if (err != cudaSuccess) {
    cout << cudaGetErrorString(err) << endl;
    assert(false);
  }
}



int main() {



  //for bear bear 
  //const int n = 128;
  //const float mu = 0.0003;
  //const float tau = .03;
  //const float nu = 1/(4*tau);
  //const float gamma = .8;
  //const int max_iter = 250000;
  //const float C = 3e1f;

  //for cat cat (first 3cases)
  //const int n = 256;
  //const float mu = 0.0001;
  //const float tau = 0.03;
  //const float nu = 1/(4*tau);
  //const float gamma = .2; 
  //const int max_iter = 1000000;
  //const float C = 3e1f;


  //for cat cat (last case)
  //const int n = 256;
  //const float mu = 0.000001;
  //const float tau = 3.0;
  //const float nu = 1/(4*tau);
  //const float gamma = .99; 
  //const int max_iter = 1000000;
  //const float C = 3e1f;


  //for spot spot
  //const int n = 32;
  //const float mu = 0.001;
  //const float tau = 0.03;
  //const float nu = 1/(4*tau);
  //const float gamma = .5; 
  //const int max_iter = 50000;
  //const float C = 1e3f;


  //for spot spot
  //const int n = 64;
  //const float mu = 0.001;
  //const float tau = 0.03;
  //const float nu = 1/(4*tau);
  //const float gamma = .5; 
  //const int max_iter = 100000;
  //const float C = 1e2f;

  //for spot spot
  const int n = 128;
  const float mu = 0.0003;
  const float tau = 0.03;
  const float nu = 1/(4*tau);
  const float gamma = .5; 
  const int max_iter = 250000;
  const float C = 1e2f;

  //for spot spot
  //const int n = 256;
  //const float mu = 0.0001;
  //const float tau = 0.03;
  //const float nu = 1/(4*tau);
  //const float gamma = .5; 
  //const int max_iter = 500000;
  //const float C = 3e1f;

  const int d = 2;
  const float dx = 4. / float(n - 1);


  //iteration data
  const int threads_per_block = 256;

  //load rho0 and rho1 from files
  float* rho0 = new float[n*n];
  float* rho1 = new float[n*n];
  rho0Init(rho0, n);
  rho1Init(rho1, n);

  float rho0_max, rho1_max;
  rho0_max = -1*std::numeric_limits<float>::infinity(); //-inf
  rho1_max = -1*std::numeric_limits<float>::infinity(); //-inf
  for (int ii=0; ii<n*n; ii++) {
    if (rho0_max < rho0[ii])
      rho0_max = rho0[ii];
    if (rho1_max < rho1[ii])
      rho1_max = rho1[ii]; 
  }
  float* t_rho0 = new float[n*n];
  float* t_rho1 = new float[n*n];
  float* t_rho0_temp = new float[n*n];
  float* t_rho1_temp = new float[n*n];  

  for (int ii=0; ii<n*n; ii++) {
    t_rho0[ii] = rho0[ii];
    t_rho1[ii] = rho1[ii];
    //t_rho0[ii] = 0;
    //t_rho1[ii] = 0;
  }




  //initialize m and Phi to be 0
  float* m = new float[n*n*d];
  memset(m, 0, sizeof(float)*n*n*d);
  float* Phi = new float[n*n];
  memset(Phi, 0, sizeof(float)*n*n);
  float* m_prev = new float[n*n*d];
  float* Phi_prev = new float[n*n];
  float* t_rho0_prev = new float[n*n];
  float* t_rho1_prev = new float[n*n];

  

  //---------------------------------------------------------------------
  //start error checking
  //thread_per_block must divide number of points
  assert((n*n) % threads_per_block == 0);
  if (n*n / threads_per_block > 65535) {
    cout << "maximum block number exceeded" << endl;
    cout << "threads_per_block should be bigger" << endl;
    assert(false);
  }
  assert((n*n)%1024==0); //required for parallel reduction
  //end error checking
  //---------------------------------------------------------------------

  //create CUDA pointers
  float* d_rho0;
  float* d_rho1;
  float* d_m;
  float* d_m_temp;
  float* d_Phi;
  float* d_t_rho0;
  float* d_t_rho0_temp;
  float* d_t_rho1;
  float* d_t_rho1_temp;
  float* d_min;
  float* d_max;
  float* d_sum;

  cout << "Allocating GPU memory" << endl;
  err_chk(cudaMalloc((void**)&d_rho0, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_rho1, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_m, sizeof(float)*n*n*d));
  err_chk(cudaMalloc((void**)&d_m_temp, sizeof(float)*n*n*d));
  err_chk(cudaMalloc((void**)&d_Phi, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_t_rho0, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_t_rho0_temp, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_t_rho1, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_t_rho1_temp, sizeof(float)*n*n));

  err_chk(cudaMalloc((void**)&d_min, sizeof(float)));
  err_chk(cudaMalloc((void**)&d_max, sizeof(float)));
  err_chk(cudaMalloc((void**)&d_sum, sizeof(float)));


  cout << "copying memory from host to GPU" << endl;
  err_chk(cudaMemcpy(d_m, m, sizeof(float)*n*n*d, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_Phi, Phi, sizeof(float)*n*n, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_rho0, rho0, sizeof(float)*n*n, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_rho1, rho1, sizeof(float)*n*n, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_t_rho0, t_rho0, sizeof(float)*n*n, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_t_rho1, t_rho1, sizeof(float)*n*n, cudaMemcpyHostToDevice));




float theta_mid_prev0, theta_mid_prev1;


  cout << "starting iteration" << endl;
  clock_t begin = clock();
  for (int k = 0; k < max_iter; k++) {
    //cout << k << endl;
    //m_temp = m
    //m = S(m+mu*gradPhi)
    //m_temp = 2m-m_temp
    mUpdate << < n*n / threads_per_block, threads_per_block >> > (d_Phi, d_m, d_m_temp, n, dx, mu);




  const float epsilon = 1e-6f;
  
  rhoStep << < n*n / threads_per_block, threads_per_block >> > (d_Phi, d_t_rho0, d_t_rho0_temp, d_t_rho1, d_t_rho1_temp, nu, n);
  
  
  float theta_min, theta_max, theta_mid;
  float sum_min, sum_max;
  if (k==0) {
    theta_min = std::numeric_limits<float>::infinity(); //inf
    theta_max = -1*std::numeric_limits<float>::infinity(); //-inf
    err_chk(cudaMemcpy(d_min, &theta_min, sizeof(float), cudaMemcpyHostToDevice));
    err_chk(cudaMemcpy(d_max, &theta_max, sizeof(float), cudaMemcpyHostToDevice));
    cuda_parallel_minmax <<<(n*n/1024),1024>>> (d_t_rho0, n*n, d_max, d_min);
    err_chk(cudaMemcpy(&theta_max, d_max, sizeof(float), cudaMemcpyDeviceToHost));
    err_chk(cudaMemcpy(&theta_min, d_min, sizeof(float), cudaMemcpyDeviceToHost));
    theta_min -= rho0_max;

    sum_min = gamma;
    sum_max = 0.0f;

  }
  else {

    float sum = 0.0f;
    err_chk(cudaMemcpy(d_sum, &sum, sizeof(float), cudaMemcpyHostToDevice));
    parallel_sum <<<(n*n/1024),1024>>> (d_rho0, d_t_rho0, n*n, theta_mid_prev0 - C*epsilon*abs(theta_mid_prev0), d_sum);
    err_chk(cudaMemcpy(&sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost));
    if (sum > 0.99*gamma && sum < 1.01*gamma) {
      theta_min = theta_max = theta_mid_prev1 - C*epsilon*abs(theta_mid_prev1);
    }
    else if (sum < gamma){
      theta_max = theta_mid_prev0 - C*epsilon*abs(theta_mid_prev0);
      sum_max = sum;
      float temp_min = std::numeric_limits<float>::infinity(); //inf
      err_chk(cudaMemcpy(d_min, &temp_min, sizeof(float), cudaMemcpyHostToDevice));
      cuda_parallel_min <<<(n*n/1024),1024>>> (d_t_rho0, n*n, d_min);
      err_chk(cudaMemcpy(&temp_min, d_min, sizeof(float), cudaMemcpyDeviceToHost));
      theta_min = temp_min - rho0_max;
      theta_min -= 0.1*abs(theta_min); //0.1 is to avoid numerical precision issues
      sum_min = gamma;
      //cout << "guess failed1 sum is " << sum << endl;
    }
    else {
      theta_min = theta_mid_prev0 - C*epsilon*abs(theta_mid_prev0);
      sum_min = sum;

      sum = 0.0f;
      err_chk(cudaMemcpy(d_sum, &sum, sizeof(float), cudaMemcpyHostToDevice));
      parallel_sum <<<(n*n/1024),1024>>> (d_rho0, d_t_rho0, n*n, theta_mid_prev0 + C*epsilon*abs(theta_mid_prev0), d_sum);
      err_chk(cudaMemcpy(&sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost));

      if (sum<=gamma) {
        theta_max = theta_mid_prev0 + C*epsilon*abs(theta_mid_prev0);
	sum_max = sum;
//	cout << "guess 0 succeed" << endl;
      }
      else {
        float temp_max = -1.0f*std::numeric_limits<float>::infinity(); //-inf
        err_chk(cudaMemcpy(d_max, &temp_max, sizeof(float), cudaMemcpyHostToDevice));
        cuda_parallel_max <<<(n*n/1024),1024>>> (d_t_rho0, n*n, d_max);
        err_chk(cudaMemcpy(&temp_max, d_max, sizeof(float), cudaMemcpyDeviceToHost));
        theta_max = temp_max;
	sum_max = 0.0f;
	//cout << "guess failed2 sum is " << sum << endl;
      }
    }

  }
  assert(theta_max>=theta_min);
  //cout << "interval computed" << endl;

  theta_mid = (theta_max+theta_min)/2;
  //cout << "counter 0 start" << endl;
  do {
    //cout << " counter 0" << endl;
    float sum = 0.0f;
    err_chk(cudaMemcpy(d_sum, &sum, sizeof(float), cudaMemcpyHostToDevice));
    parallel_sum <<<(n*n/1024),1024>>> (d_rho0, d_t_rho0, n*n, theta_mid, d_sum);
    err_chk(cudaMemcpy(&sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost));

    if (sum>gamma) {
      theta_min = theta_mid;
      sum_min = sum;
    }
    else {
      theta_max = theta_mid;
      sum_max = sum;
    }
    theta_mid = (theta_max+theta_min)/2;
   }  while((theta_max-theta_min)>epsilon* abs(theta_mid) && (sum_min-sum_max)>0.001 );
  theta_mid_prev0 = theta_mid;

  rho_commit << < n*n / threads_per_block, threads_per_block >> > (d_rho0, d_t_rho0, d_t_rho0_temp, n*n, theta_mid);


  //t_rho1 projection
  if (k==0) {
    theta_min = std::numeric_limits<float>::infinity(); //inf
    theta_max = -1.0f*std::numeric_limits<float>::infinity(); //-inf
    err_chk(cudaMemcpy(d_min, &theta_min, sizeof(float), cudaMemcpyHostToDevice));
    err_chk(cudaMemcpy(d_max, &theta_max, sizeof(float), cudaMemcpyHostToDevice));
    cuda_parallel_minmax <<<(n*n/1024),1024>>> (d_t_rho1, n*n, d_max, d_min);
    err_chk(cudaMemcpy(&theta_max, d_max, sizeof(float), cudaMemcpyDeviceToHost));
    err_chk(cudaMemcpy(&theta_min, d_min, sizeof(float), cudaMemcpyDeviceToHost));
    theta_min -= rho1_max;
    sum_min = gamma;
    sum_max = 0.0f;
  }
  else {

    float sum = 0.0f;
    err_chk(cudaMemcpy(d_sum, &sum, sizeof(float), cudaMemcpyHostToDevice));
    parallel_sum <<<(n*n/1024),1024>>> (d_rho1, d_t_rho1, n*n, theta_mid_prev1 - C*epsilon*abs(theta_mid_prev1), d_sum);
    err_chk(cudaMemcpy(&sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost));
    if (sum > 0.999*gamma && sum < 1.001*gamma) {
      theta_min = theta_max = theta_mid_prev1 - C*epsilon*abs(theta_mid_prev1);
    }
    else if (sum < gamma){

      theta_max = theta_mid_prev1 - C*epsilon*abs(theta_mid_prev1);
      sum_max = sum;
      
      float temp_min = std::numeric_limits<float>::infinity(); //inf
      err_chk(cudaMemcpy(d_min, &temp_min, sizeof(float), cudaMemcpyHostToDevice));
      cuda_parallel_min <<<(n*n/1024),1024>>> (d_t_rho1, n*n, d_min);
      err_chk(cudaMemcpy(&temp_min, d_min, sizeof(float), cudaMemcpyDeviceToHost));
      theta_min = temp_min - rho1_max;
      theta_min -= 0.1*abs(theta_min); //1.1 is to avoid numerical precision issues
      sum_min = gamma;
    }
    else {
      theta_min = theta_mid_prev1 - C*epsilon*abs(theta_mid_prev1);
      sum_min = sum;

	sum = 0.0f;
      err_chk(cudaMemcpy(d_sum, &sum, sizeof(float), cudaMemcpyHostToDevice));
      parallel_sum <<<(n*n/1024),1024>>> (d_rho1, d_t_rho1, n*n, theta_mid_prev1 + C*epsilon*abs(theta_mid_prev1), d_sum);
      err_chk(cudaMemcpy(&sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost));
      if (sum<=gamma) {
        theta_max = theta_mid_prev1 + C*epsilon*abs(theta_mid_prev1);
	sum_max = sum;
//	cout << "guess 1 succeed" << endl;
      }
      else {
        float temp_max = -1.0f*std::numeric_limits<float>::infinity(); //-inf
        err_chk(cudaMemcpy(d_max, &temp_max, sizeof(float), cudaMemcpyHostToDevice));
        cuda_parallel_max <<<(n*n/1024),1024>>> (d_t_rho1, n*n, d_max);
        err_chk(cudaMemcpy(&temp_max, d_max, sizeof(float), cudaMemcpyDeviceToHost));
        theta_max = temp_max;
	sum_max = 0.0f;
      }
    }

  }
  assert(theta_max>= theta_min);

  theta_mid = (theta_max+theta_min)/2;
  do {
    //cout << "counter 1" << endl;
    float sum = 0.0f;
    err_chk(cudaMemcpy(d_sum, &sum, sizeof(float), cudaMemcpyHostToDevice));
    parallel_sum <<<(n*n/1024),1024>>> (d_rho1, d_t_rho1, n*n, theta_mid, d_sum);
    err_chk(cudaMemcpy(&sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost));

    if (sum>gamma) {
      theta_min = theta_mid;
      sum_min = sum;
    }
    else {
      theta_max = theta_mid;
      sum_max = sum;
    }
    theta_mid = (theta_max+theta_min)/2;
   }  while((theta_max-theta_min)>epsilon* abs(theta_mid) && (sum_min-sum_max)>0.001 );
  theta_mid_prev1 = theta_mid;
  rho_commit << < n*n / threads_per_block, threads_per_block >> > (d_rho1, d_t_rho1, d_t_rho1_temp, n*n, theta_mid);

//  rho_commit << < n*n / threads_per_block, threads_per_block >> > (d_rho1, d_t_rho1, d_t_rho1_temp, n*n, 0);


    //Phi = Phi + tau*(div m_temp-t_rho0_temp+t_rho1_temp)
    PhiUpdate << < n*n / threads_per_block, threads_per_block >> >(d_Phi, d_m_temp, d_t_rho0_temp, d_t_rho1_temp, n, dx, tau);





    err_chk(cudaGetLastError());

  }
  cudaDeviceSynchronize();
  float runtime = float(clock() - begin) / CLOCKS_PER_SEC;
  cout << "Total runtime is " << runtime << "s." << endl;
  cout << "This is " << runtime / float(max_iter) * 1000 << "ms per iteration" << endl;


  cout << "copying m and Phi from GPU to host" << endl;
  err_chk(cudaMemcpy(m, d_m, sizeof(float)*n*n*d, cudaMemcpyDeviceToHost));
  err_chk(cudaMemcpy(Phi, d_Phi, sizeof(float)*n*n, cudaMemcpyDeviceToHost));
  err_chk(cudaMemcpy(t_rho0, d_t_rho0, sizeof(float)*n*n, cudaMemcpyDeviceToHost));
  err_chk(cudaMemcpy(t_rho1, d_t_rho1, sizeof(float)*n*n, cudaMemcpyDeviceToHost));

  float ss= 0;
  for (int ii=0; ii<n*n; ii++) {
  ss += t_rho0[ii];
  }
  cout << " total mass of t_rho0 is " << ss << endl;

  cout << "output m to file" << endl;
  writem(m, n);
  cout << "output Phi to file" << endl;
  writePhi(Phi, n);
  cout << "output t_rho0 to file" << endl;
  writerho0(t_rho0, n);
  cout << "output t_rho1 to file" << endl;
  writerho1(t_rho1, n);


  //computing the EMD
  float emd = 0;
  float temp;
  for (int ii = 0; ii < n*n; ii++) {
    temp = 0;
    for (int v = 0; v < d; v++) {
      temp += m[ii + n*n*v] * m[ii + n*n*v];
    }
    emd += sqrt(temp);
  }
  cout << "The EMD is " << emd << "." << endl;



  cout << "freeing CUDA resources" << endl;
  cudaFree(d_rho0);
  cudaFree(d_rho1);
  cudaFree(d_m);
  cudaFree(d_m_temp);
  cudaFree(d_Phi);
  cudaFree(d_t_rho0);
  cudaFree(d_t_rho0_temp);
  cudaFree(d_t_rho1);
  cudaFree(d_t_rho1_temp);
  cudaFree(d_min);
  cudaFree(d_max);
  cudaFree(d_sum);


  cout << "freeing host resources" << endl;
  delete[] rho0;
  delete[] rho1;
  delete[] m;
  delete[] Phi;
  delete[] m_prev;
  delete[] Phi_prev;
  delete[] t_rho0;
  delete[] t_rho1;
  delete[] t_rho0_temp;
  delete[] t_rho1_temp;
  delete[] t_rho0_prev;
  delete[] t_rho1_prev;

}




__device__ void fatomicMax(float* max, float val) {
  //there is no atomic add, so we use atomicCAS
  unsigned int* address_as_ui = (unsigned int*) max;
  unsigned int old = *address_as_ui;
  unsigned int assumed;

  do {
    assumed = old;
    old = atomicCAS(address_as_ui, assumed,
     __float_as_int( fmaxf(val, __int_as_float(assumed)) ));

  } while(assumed != old);
}

__device__ void fatomicMin(float* min, float val) {
  //there is no atomic add, so we use atomicCAS
  unsigned int* address_as_ui = (unsigned int*) min;
  unsigned int old = *address_as_ui;
  unsigned int assumed;

  do {
    assumed = old;
    old = atomicCAS(address_as_ui, assumed,
     __float_as_int( fminf(val, __int_as_float(assumed))) );

  } while(assumed != old);
}


//blocksize must be 1024
// num_elemts must be a multiple of 1024
__global__ void cuda_parallel_minmax(float *rho, int num_elements, float *max, float *min) {


  //Holds intermediates in shared memory reduction
  __syncthreads();

  __shared__ float buffer[1024/WARP_SIZE];

  int globalIdx = blockIdx.x * blockDim.x + threadIdx.x;
  int lane = threadIdx.x % WARP_SIZE;


  float temp;
  //while(globalIdx < num_elements) {
    // All threads in a block of 1024 take an element
    temp = rho[globalIdx];

    // All warps in this block (32) compute the sum of all
    // threads in their warp
    for(int delta = WARP_SIZE/2; delta > 0; delta /= 2) {
      temp = fmaxf(temp, __shfl_xor(temp, delta));
    }

    // Write all 32 of these partial sums to shared memory
    if(lane == 0) {
      buffer[threadIdx.x / WARP_SIZE] = temp;
    }
    __syncthreads();


    // Add the remaining 32 partial sums using a single warp
    if(threadIdx.x < WARP_SIZE) {
      temp = buffer[threadIdx.x];
      for(int delta = WARP_SIZE / 2; delta > 0; delta /= 2) {
        temp = fmaxf(temp, __shfl_xor(temp, delta));
      }
    }

    // Add this block's sum to the total sum
    if(threadIdx.x == 0) {
      fatomicMax(max, temp);
    }

    // Jump ahead 1024 * blocknumbers to the next region of numbers to sum
    //globalIdx += blockDim.x * gridDim.x;
    //__syncthreads();

  //}
  __syncthreads();
  //globalIdx = blockIdx.x * blockDim.x + threadIdx.x;
  //while(globalIdx < num_elements) {
    // All threads in a block of 1024 take an element
    temp = rho[globalIdx];

    // All warps in this block (32) compute the sum of all
    // threads in their warp
    for(int delta = WARP_SIZE/2; delta > 0; delta /= 2) {
      temp = fminf(temp, __shfl_xor(temp, delta));
    }

    // Write all 32 of these partial sums to shared memory
    if(lane == 0) {
      buffer[threadIdx.x / WARP_SIZE] = temp;
    }
    __syncthreads();


    // Add the remaining 32 partial sums using a single warp
    if(threadIdx.x < WARP_SIZE) {
      temp = buffer[threadIdx.x];
      for(int delta = WARP_SIZE / 2; delta > 0; delta /= 2) {
        temp = fminf(temp, __shfl_xor(temp, delta));
      }
    }

    // Add this block's sum to the total sum
    if(threadIdx.x == 0) {
      fatomicMin(min, temp);
    }

    // Jump ahead 1024 * blocknumbers to the next region of numbers to sum
    //globalIdx += blockDim.x * gridDim.x;
    //__syncthreads();

  //}

}




//blocksize must be 1024
// num_elemts must be a multiple of 1024
__global__ void cuda_parallel_min(float *rho, int num_elements, float *min) {


  //Holds intermediates in shared memory reduction
  __syncthreads();

  __shared__ float buffer[1024/WARP_SIZE];

  int globalIdx = blockIdx.x * blockDim.x + threadIdx.x;
  int lane = threadIdx.x % WARP_SIZE;
  
  float temp;
  //while(globalIdx < num_elements) {
    // All threads in a block of 1024 take an element
    temp = rho[globalIdx];

    // All warps in this block (32) compute the sum of all
    // threads in their warp
    for(int delta = WARP_SIZE/2; delta > 0; delta /= 2) {
      temp = fminf(temp, __shfl_xor(temp, delta));
    }

    // Write all 32 of these partial sums to shared memory
    if(lane == 0) {
      buffer[threadIdx.x / WARP_SIZE] = temp;
    }
    __syncthreads();


    // Add the remaining 32 partial sums using a single warp
    if(threadIdx.x < WARP_SIZE) {
      temp = buffer[threadIdx.x];
      for(int delta = WARP_SIZE / 2; delta > 0; delta /= 2) {
        temp = fminf(temp, __shfl_xor(temp, delta));
      }
    }

    // Add this block's sum to the total sum
    if(threadIdx.x == 0) {
      fatomicMin(min, temp);
    }

    // Jump ahead 1024 * blocknumbers to the next region of numbers to sum
    //globalIdx += blockDim.x * gridDim.x;
    //__syncthreads();

  //}
}


//blocksize must be 1024
// num_elemts must be a multiple of 1024
__global__ void cuda_parallel_max(float *rho, int num_elements, float *max) {


  //Holds intermediates in shared memory reduction
  __syncthreads();

  __shared__ float buffer[1024/WARP_SIZE];

  int globalIdx = blockIdx.x * blockDim.x + threadIdx.x;
  int lane = threadIdx.x % WARP_SIZE;


  float temp;
  //while(globalIdx < num_elements) {
    // All threads in a block of 1024 take an element
    temp = rho[globalIdx];

    // All warps in this block (32) compute the sum of all
    // threads in their warp
    for(int delta = WARP_SIZE/2; delta > 0; delta /= 2) {
      temp = fmaxf(temp, __shfl_xor(temp, delta));
    }

    // Write all 32 of these partial sums to shared memory
    if(lane == 0) {
      buffer[threadIdx.x / WARP_SIZE] = temp;
    }
    __syncthreads();


    // Add the remaining 32 partial sums using a single warp
    if(threadIdx.x < WARP_SIZE) {
      temp = buffer[threadIdx.x];
      for(int delta = WARP_SIZE / 2; delta > 0; delta /= 2) {
        temp = fmaxf(temp, __shfl_xor(temp, delta));
      }
    }

    // Add this block's sum to the total sum
    if(threadIdx.x == 0) {
      fatomicMax(max, temp);
    }

    // Jump ahead 1024 * blocknumbers to the next region of numbers to sum
    //globalIdx += blockDim.x * gridDim.x;
    //__syncthreads();

}





__global__ void parallel_sum(float *rho, float* t_rho, int num_elements, float theta_mid, float *sum) {
  //Holds intermediates in shared memory reduction
  __syncthreads();

  __shared__ float buffer[1024/WARP_SIZE];
  int globalIdx = blockIdx.x * blockDim.x + threadIdx.x;
  int lane = threadIdx.x % WARP_SIZE;


  float temp;
  while(globalIdx < num_elements) {
    // All threads in a block of 1024 take an element
    temp = fminf(rho[globalIdx], fmaxf(t_rho[globalIdx]-theta_mid,0.0f));
    
    // All warps in this block (32) compute the sum of all
    // threads in their warp
    for(int delta = WARP_SIZE/2; delta > 0; delta /= 2) {
      temp += __shfl_xor(temp, delta);
    }

    // Write all 32 of these partial sums to shared memory
    if(lane == 0) {
      buffer[threadIdx.x / WARP_SIZE] = temp;
    }
    __syncthreads();


    // Add the remaining 32 partial sums using a single warp
    if(threadIdx.x < WARP_SIZE) {
      temp = buffer[threadIdx.x];
      for(int delta = WARP_SIZE / 2; delta > 0; delta /= 2) {
        temp += __shfl_xor(temp, delta);
      }
    }

    // Add this block's sum to the total sum
    if(threadIdx.x == 0) {
      atomicAdd(sum, temp);
    }

    // Jump ahead 1024 * blocknumbers to the next region of numbers to sum
    globalIdx += blockDim.x * gridDim.x;
    __syncthreads();

  }

}






__global__ void rho_commit(float *rho, float* t_rho, float* t_rho_temp, int num_elements, float theta_mid) {
  int ii = blockDim.x*blockIdx.x + threadIdx.x;
  t_rho[ii] = fminf(rho[ii], fmaxf(t_rho[ii]-theta_mid,0.0f));
  t_rho_temp[ii] = 2*t_rho[ii] - t_rho_temp[ii];  
}



__global__ void rhoStep(float* d_Phi, float* d_t_rho0, float* d_t_rho0_temp, float* d_t_rho1, float* d_t_rho1_temp, int nu, int n) {
  //int N = n*n;
  int ii = blockDim.x*blockIdx.x + threadIdx.x;
  //int i = ii % n;
  //int j = ii / n;

  d_t_rho0_temp[ii] = d_t_rho0[ii];
  d_t_rho0[ii] += nu*d_Phi[ii];

  d_t_rho1_temp[ii] = d_t_rho1[ii];
  d_t_rho1[ii] -= nu*d_Phi[ii];
}



__global__ void mUpdate(float* d_Phi, float* d_m, float* d_m_temp, int n, float dx, float mu) {
  int N = n*n;
  int ii = blockDim.x*blockIdx.x + threadIdx.x;
  int i = ii % n;
  int j = ii / n;

  float local_d_m_x = d_m[ii + N * 0];
  float local_d_m_y = d_m[ii + N * 1];
  float local_d_m_temp_x;
  float local_d_m_temp_y;

  //m_temp = m;
  local_d_m_temp_x = local_d_m_x;
  local_d_m_temp_y = local_d_m_y;


  //m_i = m_i + \mu \nabla \Phi
  if (i < n - 1)
    local_d_m_x += mu * (d_Phi[(i + 1) + n*j] - d_Phi[ii]) / dx;
  // else
  // local_d_m_x = 0;

  if (j < n - 1)
    local_d_m_y += mu * (d_Phi[i + n*(j + 1)] - d_Phi[ii]) / dx;
  // else
  // local_d_m_y = 0;


  /*
  //Shrink1
  float eps = 0.001;
  float shrink_factor;
  if (abs(local_d_m_x) <= mu)
  shrink_factor = 0;
  else
  shrink_factor = (1 - mu / abs(local_d_m_x))/(1 + mu*eps);
  local_d_m_x *= shrink_factor;
  if (abs(local_d_m_y) <= mu)
  shrink_factor = 0;
  else
  shrink_factor = (1 - mu / abs(local_d_m_y))/(1 + mu*eps);
  local_d_m_y *= shrink_factor;
  */

  //Shrink2
  float shrink_factor, norm;
  norm = sqrt(local_d_m_x*local_d_m_x + local_d_m_y*local_d_m_y);
  if (norm <= mu)
    shrink_factor = 0.0f;
  else
    shrink_factor = 1 - mu / norm;
  local_d_m_x *= shrink_factor;
  local_d_m_y *= shrink_factor;


  d_m[ii + N * 0] = local_d_m_x;
  d_m[ii + N * 1] = local_d_m_y;

  //m_temp = 2m - m_temp
  d_m_temp[ii + N * 0] = 2.0f * local_d_m_x - local_d_m_temp_x;
  d_m_temp[ii + N * 1] = 2.0f * local_d_m_y - local_d_m_temp_y;
}


__global__ void PhiUpdate(float* d_Phi, float *d_m_temp, float* d_t_rho0_temp, float* d_t_rho1_temp, int n, float dx, float tau) {
  int N = n*n;
  int ii = blockDim.x*blockIdx.x + threadIdx.x;
  int i = ii % n;
  int j = (ii / n);

  //divm = divergence * m_temp
  float m_minus;
  float divm = 0;

  //x-gradient on m_x
  if (i >= 1)
    m_minus = d_m_temp[(i - 1) + n*j + N * 0];
  else
    m_minus = 0;
  divm += (d_m_temp[ii + N * 0] - m_minus) / dx;

  //y-gradient on m_y
  if (j >= 1)
    m_minus = d_m_temp[i + n*(j - 1) + N * 1];
  else
    m_minus = 0;
  divm += (d_m_temp[ii + N * 1] - m_minus) / dx;


  //Phi = Phi + tau ( divergence * m_temp + rho1-temp - rho0_temp )
  d_Phi[ii] += tau * (divm + d_t_rho1_temp[ii] - d_t_rho0_temp[ii]);

}

